"""Data classes for GMM"""
from __future__ import annotations

import os
import typing

from _kalpy.matrix import DoubleVector
from _kalpy.util import (
    RandomAccessBaseDoubleVectorReader,
    RandomAccessInt32VectorVectorReader,
    SequentialBaseDoubleVectorReader,
    SequentialInt32VectorVectorReader,
)
from kalpy.data import PathLike
from kalpy.utils import generate_read_specifier, read_kaldi_object


class IvectorArchive:
    """
    Class for reading an archive or SCP of ivector

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(self, file_name: PathLike, num_utterances_file_name: PathLike = None):
        if not os.path.exists(file_name):
            raise OSError(f"Specified file does not exist: {file_name}")
        self.file_name = str(file_name)
        self.num_utterances_file_name = num_utterances_file_name
        self.read_specifier = generate_read_specifier(file_name)
        self.random_reader = RandomAccessBaseDoubleVectorReader(self.read_specifier)
        self.num_utterances_mapping = {}
        if self.num_utterances_file_name is not None:
            with open(self.num_utterances_file_name) as f:
                for line in f:
                    line = line.strip()
                    speaker_id, num_utts = line.split()
                    self.num_utterances_mapping[speaker_id] = int(num_utts)

    def close(self):
        if self.random_reader.IsOpen():
            self.random_reader.Close()

    @property
    def sequential_reader(self) -> SequentialBaseDoubleVectorReader:
        """Sequential reader for lattices"""
        return SequentialBaseDoubleVectorReader(self.read_specifier)

    def __iter__(self) -> typing.Generator[typing.Tuple[str, DoubleVector]]:
        """Iterate over the utterance lattices in the archive"""
        if self.read_specifier.startswith("scp"):
            with open(self.file_name, encoding="utf8") as f:
                for line in f:
                    line = line.strip()
                    key, ark_path = line.split(maxsplit=1)
                    ivector = read_kaldi_object(DoubleVector, ark_path)
                    num_utterances = self.num_utterances_mapping.get(key, 1)
                    yield key, ivector, num_utterances
        else:
            reader = self.sequential_reader
            try:
                while not reader.Done():
                    key = reader.Key()
                    ivector = reader.Value()
                    num_utterances = self.num_utterances_mapping.get(key, 1)
                    yield key, ivector, num_utterances
                    reader.Next()
            finally:
                reader.Close()

    def __del__(self):
        self.close()

    def __getitem__(self, item: str) -> DoubleVector:
        """Get lattice for a particular key from the archive file"""
        item = str(item)
        if not self.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.file_name}")
        return self.random_reader.Value(item)


class GselectArchive:
    """
    Class for reading an archive or SCP of gaussian selections

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(self, file_name: PathLike):
        if not os.path.exists(file_name):
            raise OSError(f"Specified file does not exist: {file_name}")
        self.file_name = str(file_name)
        self.read_specifier = generate_read_specifier(file_name)
        self.random_reader = RandomAccessInt32VectorVectorReader(self.read_specifier)

    def close(self):
        if self.random_reader.IsOpen():
            self.random_reader.Close()

    @property
    def sequential_reader(self) -> SequentialInt32VectorVectorReader:
        """Sequential reader for lattices"""
        return SequentialInt32VectorVectorReader(self.read_specifier)

    def __iter__(self) -> typing.Generator[typing.Tuple[str, typing.List[typing.List[int]]]]:
        """Iterate over the utterance lattices in the archive"""
        reader = self.sequential_reader
        try:
            while not reader.Done():
                utt = reader.Key()
                value = reader.Value()
                yield utt, value
                reader.Next()
        finally:
            reader.Close()

    def __del__(self):
        self.close()

    def __getitem__(self, item: str) -> typing.List[typing.List[int]]:
        """Get lattice for a particular key from the archive file"""
        item = str(item)
        if not self.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.file_name}")
        return self.random_reader.Value(item)



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import pynini

    from kalpy.gmm.data import CtmInterval, WordCtmInterval


class KalpyError(Exception):
    """
    Base exception class
    """

    def __init__(self, base_error_message: str, *args, **kwargs):
        self.message_lines: typing.List[str] = [base_error_message]

    @property
    def message(self) -> str:
        """Formatted exception message"""
        return "\n".join(self.message_lines)

    def __str__(self) -> str:
        """Output the error"""
        message = type(self).__name__ + ":"
        message += "\n\n" + self.message
        return message


class CtmError(KalpyError):
    """
    Class for errors in creating CTM intervals

    Parameters
    ----------
    ctm: :class:`~kalpy.gmm.data.CtmInterval`
        CTM interval that was not parsed correctly
    """

    def __init__(self, ctm: typing.Union[CtmInterval, WordCtmInterval]):
        KalpyError.__init__(self, f"Error was encountered in processing CTM interval: {ctm}")


class ReadError(KalpyError):
    """
    Class for reading files
    """

    pass


class AlignerError(KalpyError):
    """
    Class for errors during alignment
    """

    pass


class AcousticModelError(KalpyError):
    pass


class LexiconError(KalpyError):
    pass


class PhonesToPronunciationsError(KalpyError):
    """
    Class for errors in creating pronunciations from phones

    Parameters
    ----------
    ctm: :class:`~kalpy.gmm.data.CtmInterval`
        CTM interval that was not parsed correctly
    """

    def __init__(self, text: str, phones: str, phone_fst: pynini.Fst, phone_to_word: pynini.Fst):
        KalpyError.__init__(self, f"Error was encountered in creating pronunciations for: {text}")
        self.message_lines.append(f"Phones: {phones}")
        self.message_lines.append("Phone FST:")
        self.message_lines.append(str(phone_fst))
        self.message_lines.append("Phone to word FST:")
        self.message_lines.append(str(phone_to_word))



"""Data classes for kalpy"""
from __future__ import annotations

import os.path
import pathlib
import typing
from dataclasses import dataclass

import librosa
import numpy as np

from _kalpy.matrix import FloatMatrix
from _kalpy.util import (
    RandomAccessBaseDoubleMatrixReader,
    RandomAccessBaseFloatMatrixReader,
    SequentialBaseDoubleMatrixReader,
    SequentialBaseFloatMatrixReader,
)
from kalpy.utils import generate_read_specifier

PathLike = typing.Union[str, pathlib.Path]


@dataclass
class Segment:
    """
    Data class for information about acoustic segments
    """

    file_path: str
    begin: typing.Optional[float] = 0.0
    end: typing.Optional[float] = None
    channel: typing.Optional[int] = 0

    def load_audio(self):
        duration = None
        if self.end is not None and self.begin is not None:
            duration = self.end - self.begin
        y, _ = librosa.load(
            self.file_path,
            sr=16000,
            offset=self.begin,
            duration=duration,
            mono=False,
        )
        if len(y.shape) > 1:
            channel = 0 if self.channel is None else self.channel
            y = y[channel, :]
        if duration is None:
            self.end = y.shape[0] / 16_000
        return y

    @property
    def wave(self):
        if getattr(self, "_wave", None) is None:
            self._wave = self.load_audio()
        return self._wave

    @property
    def kaldi_wave(self):
        return np.round(self.wave * 32768)


def make_scp_safe(string: str) -> str:
    """
    Helper function to make a string safe for saving in Kaldi scp files.  They use space as a delimiter, so
    any spaces in the string will be converted to "_KALDISPACE_" to preserve them

    Parameters
    ----------
    string: str
        Text to escape

    Returns
    -------
    str
        Escaped text
    """
    return str(string).replace(" ", "_KALDISPACE_")


def load_scp_safe(string: str) -> str:
    """
    Helper function to load previously made safe text.  All instances of "_KALDISPACE_" will be converted to a
    regular space character

    Parameters
    ----------
    string: str
        String to convert

    Returns
    -------
    str
        Converted string
    """
    return string.replace("_KALDISPACE_", " ")


class KaldiMapping(dict):
    def __init__(self, *args, list_mapping: bool = False, **kwargs):
        self.list_mapping = list_mapping
        super().__init__(*args, **kwargs)

    def export(self, file_path: typing.Union[str, pathlib.Path], skip_safe: bool = False) -> None:
        with open(file_path, "w", encoding="utf8") as f:
            for k in sorted(self.keys()):
                v = self[k]
                if isinstance(v, (list, set, tuple)):
                    v = " ".join(map(str, v))
                elif not skip_safe:
                    v = make_scp_safe(v)
                f.write(f"{make_scp_safe(k)} {v}\n")

    def load(self, file_path, data_type: typing.Optional[typing.Type] = str):
        self.clear()
        with open(file_path, "r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if line == "":
                    continue
                line_list = line.split()
                key = load_scp_safe(line_list.pop(0))
                if len(line_list) == 1 and not self.list_mapping:
                    value = data_type(line_list[0])
                    if isinstance(value, str):
                        value = load_scp_safe(value)
                else:
                    value = [data_type(x) for x in line_list if x not in ["[", "]"]]
                self[key] = value


class MatrixArchive:
    """
    Class for reading an archive or SCP of matrices

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(self, file_name: typing.Union[pathlib.Path, str], double: bool = False):
        if not os.path.exists(file_name):
            raise OSError(f"Specified file does not exist: {file_name}")
        self.file_name = str(file_name)
        self.double = double
        self.read_specifier = generate_read_specifier(file_name)
        if self.double:
            self.random_reader = RandomAccessBaseDoubleMatrixReader(self.read_specifier)
        else:
            self.random_reader = RandomAccessBaseFloatMatrixReader(self.read_specifier)

    def close(self):
        if getattr(self, "random_reader", None) is not None and self.random_reader.IsOpen():
            self.random_reader.Close()

    def __del__(self):
        self.close()

    @property
    def sequential_reader(self) -> SequentialBaseFloatMatrixReader:
        if self.double:
            return SequentialBaseDoubleMatrixReader(self.read_specifier)
        return SequentialBaseFloatMatrixReader(self.read_specifier)

    def __iter__(self) -> typing.Generator[typing.Tuple[str, FloatMatrix]]:
        """Iterate over the utterance features in the archive"""
        reader = self.sequential_reader
        try:
            while not reader.Done():
                utt = reader.Key()
                feats = reader.Value()
                yield utt, feats
                reader.Next()
        finally:
            reader.Close()

    def __getitem__(self, item: str) -> FloatMatrix:
        """Get features for a particular key from the archive file"""
        item = str(item)
        if not self.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.file_name}")
        return self.random_reader.Value(item)



import os

from kalpy.data import Segment
from kalpy.feat.data import FeatureArchive
from kalpy.feat.pitch import PitchComputer


def test_generate_pitch(wav_path):
    feature_generator = PitchComputer(snip_edges=False)
    pitch = feature_generator.compute_pitch(Segment(wav_path))
    assert pitch.shape[0] == 2672
    assert pitch.shape[1] == 3
    pitch = feature_generator.compute_pitch(Segment(wav_path, 1, 2, 0))
    assert pitch.shape[0] == 100
    assert pitch.shape[1] == 3


def test_export_pitch(wav_path, temp_dir):
    output_file_name = temp_dir.joinpath("mfccs.ark")
    feature_generator = PitchComputer(snip_edges=False)
    segments = {"1-1": Segment(wav_path)}
    feature_generator.export_feats(output_file_name, segments.items())
    assert output_file_name.exists()

    archive = FeatureArchive(output_file_name)
    for utt, pitch in archive:
        pitch = pitch.numpy()
        assert utt == "1-1"
        assert pitch.shape[0] == 2672
        assert pitch.shape[1] == 3
    pitch = archive["1-1"].numpy()
    assert pitch.shape[0] == 2672
    assert pitch.shape[1] == 3

    archive.close()
    os.remove(output_file_name)
    feature_generator.export_feats(
        output_file_name, segments.items(), write_scp=True, compress=False
    )
    assert output_file_name.with_suffix(".scp").exists()

    archive = FeatureArchive(output_file_name.with_suffix(".scp"))
    for utt, pitch in archive:
        pitch = pitch.numpy()
        assert utt == "1-1"
        assert pitch.shape[0] == 2672
        assert pitch.shape[1] == 3
    pitch = archive["1-1"].numpy()
    assert pitch.shape[0] == 2672
    assert pitch.shape[1] == 3



from kalpy.aligner import KalpyAligner
from kalpy.data import Segment
from kalpy.models import AcousticModel
from kalpy.utterance import Utterance


def test_aligner_align(
    english_mfa_am_dir,
    mfa_dictionary_path,
    mfa_wav_path,
):
    am = AcousticModel(english_mfa_am_dir)
    lc = am.lexicon_compiler
    lc.load_pronunciations(mfa_dictionary_path)
    aligner = KalpyAligner(am, lc)
    seg = Segment(mfa_wav_path)
    utterance = Utterance(seg, "montreal forced aligner")
    ctm = aligner.align_utterance(utterance)
    assert len(ctm.word_intervals) == 3


def test_aligner_fine_tune(
    english_mfa_am_dir,
    mfa_dictionary_path,
    mfa_wav_path,
):
    am = AcousticModel(english_mfa_am_dir)
    lc = am.lexicon_compiler
    lc.load_pronunciations(mfa_dictionary_path)
    aligner = KalpyAligner(am, lc)
    seg = Segment(mfa_wav_path)
    utterance = Utterance(seg, "montreal forced aligner")
    ctm = aligner.align_utterance(utterance)
    assert len(ctm.word_intervals) == 3
    fine_tuned_ctm = aligner.fine_tune_alignments(utterance)
    assert len(fine_tuned_ctm.word_intervals) == 3
    boundaries = ctm.phone_boundaries
    fine_tuned_boundaries = fine_tuned_ctm.phone_boundaries
    assert len(boundaries) == len(fine_tuned_boundaries)
    for i, b in enumerate(boundaries):
        assert abs(b - fine_tuned_boundaries[i]) < 0.01



import pickle

from _kalpy.util import CtmInterval, Interval, IntervalAlignment, WordCtmInterval, align_intervals
from kalpy.evaluation import format_alignment


def test_intervals():
    i = Interval(0.0, 1.0, "hello")
    assert i.begin == 0.0
    assert i.end == 1.0
    assert i.label == "hello"
    mapping = {}
    silence_phone = "sil"
    silence_phones = {silence_phone}
    assert i.compare_labels(silence_phone, silence_phones, mapping) == 10
    assert i.compare_labels("hello", silence_phones, mapping) == 0.0
    assert i.compare_labels("other_hello", silence_phones, mapping) == 2.0

    mapping = {"other_hello": {"hello"}}
    assert i.compare_labels("other_hello", silence_phones, mapping) == 0.0

    i_data = pickle.dumps(i)
    i_new = pickle.loads(i_data)
    assert i_new.begin == i.begin
    assert i_new.end == i.end
    assert i_new.label == i.label


def test_ctm_intervals():
    i = CtmInterval(0.0, 1.0, "hello", 0)
    assert i.begin == 0.0
    assert i.end == 1.0
    assert i.label == "hello"
    assert i.symbol == 0
    mapping = {}
    silence_phone = "sil"
    silence_phones = {silence_phone}
    assert i.compare_labels(silence_phone, silence_phones, mapping) == 10
    assert i.compare_labels("hello", silence_phones, mapping) == 0.0
    assert i.compare_labels("other_hello", silence_phones, mapping) == 2.0

    mapping = {"other_hello": {"hello"}}
    assert i.compare_labels("other_hello", silence_phones, mapping) == 0.0

    i_data = pickle.dumps(i)
    i_new = pickle.loads(i_data)
    assert i_new.begin == i.begin
    assert i_new.end == i.end
    assert i_new.label == i.label


def test_word_ctm_intervals():
    i = CtmInterval(0.0, 1.0, "hh", 0)
    w = WordCtmInterval("hello", 0, [i])
    assert w.begin == 0.0
    assert w.end == 1.0
    assert w.label == "hello"
    assert w.symbol == 0
    assert w.phones[0].begin == 0.0
    assert w.phones[0].end == 1.0
    assert w.phones[0].label == "hh"
    assert w.phones[0].symbol == 0
    mapping = {}
    silence_phone = "sil"
    silence_phones = {silence_phone}
    assert w.compare_labels(silence_phone, silence_phones, mapping) == 10
    assert w.compare_labels("hello", silence_phones, mapping) == 0.0
    assert w.compare_labels("other_hello", silence_phones, mapping) == 2.0

    mapping = {"other_hello": {"hello"}}
    assert w.compare_labels("other_hello", silence_phones, mapping) == 0.0

    w_data = pickle.dumps(w)
    w_new = pickle.loads(w_data)
    assert w_new.begin == w.begin
    assert w_new.end == w.end
    assert w_new.label == w.label
    assert len(w_new.phones) == 1
    assert w_new.phones[0].begin == w.phones[0].begin
    assert w_new.phones[0].end == w.phones[0].end
    assert w_new.phones[0].label == w.phones[0].label
    assert w_new.phones[0].symbol == w.phones[0].symbol


def test_align_intervals(reference_hello_intervals, test_hi_intervals):
    mapping = {}
    silence_phone = "sil"
    alignment = align_intervals(
        reference_hello_intervals, test_hi_intervals, {silence_phone}, mapping
    )
    assert len(alignment) == 4
    for r, t in alignment:
        print(r, t)
    assert alignment[0][0].label == alignment[0][1].label
    assert alignment[1][0].label != alignment[1][1].label
    assert alignment[2][1].label == alignment[3][1].label == "-"
    assert alignment[2][1].begin == alignment[3][1].begin == -1.0


def test_format_alignment(reference_hello_intervals, test_hi_intervals):
    mapping = {}
    silence_phone = "sil"
    alignment = align_intervals(
        reference_hello_intervals, test_hi_intervals, {silence_phone}, mapping
    )
    assert len(alignment) == 4
    for r, t in alignment:
        print(r, t)
    assert alignment[0][0].label == alignment[0][1].label
    assert alignment[1][0].label != alignment[1][1].label
    assert alignment[2][1].label == alignment[3][1].label == "-"
    assert alignment[2][1].begin == alignment[3][1].begin == -1.0
    formatted = format_alignment(alignment)
    print(formatted)
    assert "hh eh l ow" in formatted
    assert "   |  .      " in formatted
    assert "hh ay - -" in formatted
    assert "Score=6" in formatted

    mapping = {"ay": {"l"}}
    silence_phone = "sil"
    alignment = align_intervals(
        reference_hello_intervals, test_hi_intervals, {silence_phone}, mapping
    )
    assert len(alignment) == 4
    for r, t in alignment:
        print(r, t)
    assert alignment[0][0].label == alignment[0][1].label
    assert alignment[1][0].label == "eh"
    assert alignment[1][1].label == "-"
    assert alignment[2][0].label == "l"
    assert alignment[2][1].label == "ay"
    assert alignment[1][1].label == alignment[3][1].label == "-"
    assert alignment[1][1].begin == alignment[3][1].begin == -1.0
    formatted = format_alignment(alignment)
    print(formatted)
    assert "hh eh l  ow" in formatted
    assert "|     . " in formatted
    assert "hh -  ay -" in formatted
    assert "Score=6" in formatted



"""Classes for generating training graphs"""
import logging
import pathlib
import typing

import pynini
import pywrapfst

from _kalpy.decoder import TrainingGraphCompiler as _TrainingGraphCompiler
from _kalpy.decoder import TrainingGraphCompilerOptions
from _kalpy.fstext import (
    VectorFst,
    VectorFstWriter,
    fst_add_self_loops,
    fst_arc_sort,
    fst_compose_context,
    fst_determinize_star,
    fst_minimize_encoded,
    fst_push_special,
    fst_rm_eps_local,
    fst_rm_symbols,
    fst_table_compose,
)
from _kalpy.hmm import TransitionModel, make_h_transducer
from _kalpy.tree import ContextDependency
from _kalpy.util import ReadKaldiObject
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.fstext.utils import kaldi_to_pynini, pynini_to_kaldi
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.graphs")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class TrainingGraphCompiler:
    """
    Parameters
    ----------
    model_path: str
        Path to model file
    tree_path: str
        Path to tree file
    lexicon_compiler: :class:`~kalpy.fstext.lexicon.LexiconCompiler`
        Lexicon compiler to use in generating training graphs
    transition_scale: float
        Scale on transitions, typically set to 0 as it will be defined during alignment
    self_loop_scale: float
        Scale on transitions, typically set to 0 as it will be defined during alignment
    batch_size: int
        Batch size for compilation, larger batches will be faster but use more memory

    Attributes
    ----------
    transition_model: :class:`~_kalpy.hmm.TransitionModel`
        Transition model to use in compiling the graphs
    tree: :class:`~_kalpy.hmm.TransitionModel`
        Context dependency tree to use in compiling the graphs
    """

    def __init__(
        self,
        model_path: typing.Union[pathlib.Path, str],
        tree_path: typing.Union[pathlib.Path, str],
        lexicon_compiler: LexiconCompiler,
        transition_scale: float = 0.0,
        self_loop_scale: float = 0.0,
        batch_size: int = 1000,
        use_g2p: bool = False,
        disambiguation_symbols: typing.List[int] = None,
        oov_word: str = "<unk>",
    ):
        self.transition_model = TransitionModel()
        ReadKaldiObject(str(model_path), self.transition_model)
        self.tree = ContextDependency()
        ReadKaldiObject(str(tree_path), self.tree)
        self.batch_size = batch_size
        self.options = TrainingGraphCompilerOptions(transition_scale, self_loop_scale)
        self._compiler = None
        self.use_g2p = use_g2p
        self.lexicon_path = None
        self.lexicon_compiler = lexicon_compiler
        self.oov_word = oov_word
        if disambiguation_symbols is None:
            disambiguation_symbols = []
        self.disambiguation_symbols = disambiguation_symbols
        self._kaldi_fst = self.lexicon_compiler.fst
        if not isinstance(self._kaldi_fst, VectorFst):
            self._kaldi_fst = VectorFst.from_pynini(self._kaldi_fst)

    def __del__(self):
        del self._compiler
        del self._kaldi_fst

    @property
    def word_table(self):
        return self.lexicon_compiler.word_table

    def to_int(self, word: str) -> int:
        """
        Look up a word in the word symbol table

        Parameters
        ----------
        word: str
            Word to look up

        Returns
        -------
        int
            Integer ID of word in symbol table
        """
        if self.word_table.member(word):
            return self.word_table.find(word)
        return self.word_table.find(self.oov_word)

    @property
    def compiler(self):
        if self._compiler is None:
            disambiguation_symbols = []
            if self.lexicon_compiler is not None and self.lexicon_compiler.disambiguation:
                disambiguation_symbols = self.lexicon_compiler.disambiguation_symbols
            self._compiler = _TrainingGraphCompiler(
                self.transition_model,
                self.tree,
                self._kaldi_fst,
                disambiguation_symbols,
                self.options,
            )
        return self._compiler

    def export_graphs(
        self,
        file_name: typing.Union[pathlib.Path, str],
        transcripts: typing.Iterable[typing.Tuple[str, str]],
        write_scp: bool = False,
        callback: typing.Callable = None,
        interjection_words: typing.List[str] = None,
        cutoff_pattern: str = None,
    ):
        """
        Export training graphs to a kaldi archive file (i.e., fsts.ark)

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Archive file path to export to
        transcripts: iterable[tuple[str, str]]
            Dictionary of utterance IDs to transcripts
        write_scp: bool
            Flag for whether an SCP file should be generated as well
        callback: callable, optional
            Optional callback function for progress updates
        interjection_words: list[str], optional
            List of words to add as interjections to the transcripts
        cutoff_pattern: str, optional
            Cutoff symbol to use for inserting cutoffs before words
        """
        write_specifier = generate_write_specifier(file_name, write_scp)
        writer = VectorFstWriter(write_specifier)
        keys = []
        transcript_batch = []
        original_transcripts = []
        num_done = 0
        num_error = 0
        logger.debug(f"DISAMBIGUATION: {self.lexicon_compiler.disambiguation}")
        for key, transcript in transcripts:
            keys.append(key)
            original_transcripts.append(transcript)
            if self.use_g2p:
                transcript_batch.append(transcript)
            elif interjection_words:
                transcript_batch.append(
                    self.generate_utterance_graph(transcript, interjection_words, cutoff_pattern)
                )
            else:
                transcript_batch.append([self.to_int(x) for x in transcript.split()])
            if len(keys) >= self.batch_size:
                if self.use_g2p:
                    fsts = []
                    for t in transcript_batch:
                        fsts.append(self.compile_fst(t))
                elif interjection_words:
                    fsts = self.compiler.CompileGraphs(transcript_batch)
                else:
                    fsts = self.compiler.CompileGraphsFromText(transcript_batch)
                assert len(fsts) == len(keys)
                batch_done = 0
                batch_error = 0
                for i, key in enumerate(keys):
                    fst = fsts[i]
                    if fst.Start() == pywrapfst.NO_STATE_ID:
                        original_transcript = original_transcripts[i]
                        transcript = transcript_batch[i]
                        logger.warning(
                            f"Skipping {key}, empty FST for {transcript} ({original_transcript})"
                        )
                        batch_error += 1
                        continue
                    writer.Write(str(key), fst)
                    del fst
                    batch_done += 1
                num_done += batch_done
                num_error += batch_error
                logger.debug(f"Done {num_done} utterances, errors on {num_error}.")
                if callback:
                    callback(batch_done)
                keys = []
                transcript_batch = []
                original_transcripts = []
                del fsts
        if keys:
            if self.use_g2p:
                fsts = []
                for t in transcript_batch:
                    fsts.append(self.compile_fst(t))
            elif interjection_words:
                fsts = self.compiler.CompileGraphs(transcript_batch)
            else:
                fsts = self.compiler.CompileGraphsFromText(transcript_batch)
            del transcript_batch
            assert len(fsts) == len(keys)
            batch_done = 0
            batch_error = 0
            for i, key in enumerate(keys):
                fst = fsts[i]
                if fst.Start() == pywrapfst.NO_STATE_ID:
                    logger.warning(f"Skipping {key}, empty FST")
                    batch_error += 1
                    continue
                writer.Write(str(key), fst)
                batch_done += 1
                del fst
            num_done += batch_done
            num_error += batch_error
            del fsts
            if callback:
                callback(batch_done)
        writer.Close()
        logger.info(f"Done {num_done} utterances, errors on {num_error}.")

    def generate_utterance_graph(
        self,
        transcript: str,
        interjection_words: typing.List[str] = None,
        cutoff_pattern: str = None,
    ) -> typing.Optional[VectorFst]:
        if interjection_words is None:
            interjection_words = []
        default_interjection_cost = 3.0
        cutoff_interjection_cost = default_interjection_cost
        cutoff_symbol = -1
        if cutoff_pattern is not None and self.word_table.member(cutoff_pattern):
            cutoff_symbol = self.to_int(cutoff_pattern)
        interjection_costs = {}
        if interjection_words:
            for iw in interjection_words:
                if not self.word_table.member(iw):
                    continue
                if isinstance(interjection_words, dict):
                    interjection_cost = interjection_words[iw] * default_interjection_cost
                else:
                    interjection_cost = default_interjection_cost
                if isinstance(iw, str):
                    iw = self.to_int(iw)
                if iw == cutoff_symbol:
                    cutoff_interjection_cost = interjection_cost
                    continue
                interjection_costs[iw] = interjection_cost
        g = pynini.Fst()
        start_state = g.add_state()
        g.set_start(start_state)
        if isinstance(transcript, str):
            transcript = transcript.split()
        for word_symbol in transcript:
            if not isinstance(word_symbol, int):
                word_symbol = self.to_int(word_symbol)
            interjection_state = g.add_state()
            for iw_symbol, interjection_cost in interjection_costs.items():
                g.add_arc(
                    start_state,
                    pywrapfst.Arc(
                        iw_symbol,
                        iw_symbol,
                        pywrapfst.Weight(g.weight_type(), interjection_cost),
                        interjection_state,
                    ),
                )
            if cutoff_pattern is not None:
                cutoff_word = f"{cutoff_pattern[:-1]}-{self.word_table.find(word_symbol)}{cutoff_pattern[-1]}"
                if self.word_table.member(cutoff_word):
                    iw_symbol = self.to_int(cutoff_word)
                    g.add_arc(
                        start_state,
                        pywrapfst.Arc(
                            iw_symbol,
                            iw_symbol,
                            pywrapfst.Weight(g.weight_type(), cutoff_interjection_cost),
                            interjection_state,
                        ),
                    )
            g.add_arc(
                start_state,
                pywrapfst.Arc(
                    word_symbol,
                    word_symbol,
                    pywrapfst.Weight(g.weight_type(), default_interjection_cost),
                    interjection_state,
                ),
            )
            g.add_arc(
                start_state,
                pywrapfst.Arc(
                    self.word_table.find("<eps>"),
                    self.word_table.find("<eps>"),
                    pywrapfst.Weight(g.weight_type(), 1.0),
                    interjection_state,
                ),
            )
            final_state = g.add_state()
            g.add_arc(
                interjection_state,
                pywrapfst.Arc(
                    word_symbol,
                    word_symbol,
                    pywrapfst.Weight.one(g.weight_type()),
                    final_state,
                ),
            )
            start_state = final_state
        final_state = g.add_state()
        for iw_symbol, interjection_cost in interjection_costs.items():
            g.add_arc(
                start_state,
                pywrapfst.Arc(
                    iw_symbol,
                    iw_symbol,
                    pywrapfst.Weight(g.weight_type(), interjection_cost),
                    final_state,
                ),
            )
        g.add_arc(
            start_state,
            pywrapfst.Arc(
                self.word_table.find("<eps>"),
                self.word_table.find("<eps>"),
                pywrapfst.Weight.one(g.weight_type()),
                final_state,
            ),
        )
        g.set_final(final_state, pywrapfst.Weight.one(g.weight_type()))
        g = VectorFst.from_pynini(g)
        return g

    def compile_fst(
        self,
        transcript: str,
        interjection_words: typing.List[str] = None,
        cutoff_pattern: str = None,
    ) -> typing.Optional[VectorFst]:
        """
        Compile a transcript to a training graph

        Parameters
        ----------
        transcript: str
            Orthographic transcript to compile
        interjection_words: list[str], optional
            List of words to add as interjections to the transcript
        cutoff_pattern: str, optional
            Cutoff symbol to use for inserting cutoffs before words

        Returns
        -------
        :class:`_kalpy.fstext.VectorFst`
            Training graph of transcript
        """
        if self.use_g2p:
            g_fst = pynini.accep(transcript, token_type=self.word_table)
            lg_fst = pynini.compose(g_fst, self._fst, compose_filter="alt_sequence")
            lg_fst = lg_fst.project("output").rmepsilon()
            weight_type = lg_fst.weight_type()
            weight_threshold = pywrapfst.Weight(weight_type, 2.0)
            state_threshold = 256 + 2 * lg_fst.num_states()
            lg_fst = pynini.determinize(lg_fst, nstate=state_threshold, weight=weight_threshold)
            lg_fst = VectorFst.from_pynini(lg_fst)
            disambig_syms_in = (
                []
                if not self.lexicon_compiler.disambiguation
                else self.lexicon_compiler.disambiguation_symbols
            )
            lg_fst = fst_determinize_star(lg_fst, use_log=True)
            fst_minimize_encoded(lg_fst)
            fst_push_special(lg_fst)
            clg_fst, disambig_out, ilabels = fst_compose_context(
                lg_fst,
                disambig_syms_in,
                self.tree.ContextWidth(),
                self.tree.CentralPosition(),
            )
            fst_arc_sort(clg_fst, sort_type="ilabel")
            h, disambig = make_h_transducer(self.tree, self.transition_model, ilabels)
            fst = fst_table_compose(h, clg_fst)
            if fst.Start() == pywrapfst.NO_STATE_ID:
                logger.debug(f"Falling back to pynini compose for '{transcript}")
                h = kaldi_to_pynini(h)
                clg_fst = kaldi_to_pynini(clg_fst)
                fst = pynini_to_kaldi(pynini.compose(h, clg_fst))
            fst_determinize_star(fst, use_log=True)
            fst_rm_symbols(fst, disambig)
            fst_rm_eps_local(fst)
            fst_minimize_encoded(fst)
            fst_add_self_loops(
                fst, self.transition_model, disambig_syms_in, self.options.self_loop_scale
            )
        elif interjection_words:
            g = self.generate_utterance_graph(transcript, interjection_words, cutoff_pattern)
            # fst = VectorFst()
            # self.compiler.CompileGraph(g, fst)
            # lg_fst = pynini.compose(self._fst, g, compose_filter="alt_sequence")
            # lg_fst = VectorFst.from_pynini(lg_fst)
            lg_fst = fst_table_compose(self._kaldi_fst, g)

            disambig_syms_in = (
                []
                if not self.lexicon_compiler.disambiguation
                else self.lexicon_compiler.disambiguation_symbols
            )
            lg_fst = fst_determinize_star(lg_fst, use_log=True)
            fst_minimize_encoded(lg_fst)
            fst_push_special(lg_fst)
            clg_fst, disambig_out, ilabels = fst_compose_context(
                lg_fst,
                disambig_syms_in,
                self.tree.ContextWidth(),
                self.tree.CentralPosition(),
            )
            fst_arc_sort(clg_fst, sort_type="ilabel")
            h, disambig = make_h_transducer(self.tree, self.transition_model, ilabels)
            fst = fst_table_compose(h, clg_fst)
            if fst.Start() == pywrapfst.NO_STATE_ID:
                logger.debug(f"Falling back to pynini compose for '{transcript}")
                h = kaldi_to_pynini(h)
                clg_fst = kaldi_to_pynini(clg_fst)
                fst = pynini_to_kaldi(pynini.compose(h, clg_fst))
            fst_determinize_star(fst, use_log=True)
            fst_rm_symbols(fst, disambig)
            fst_rm_eps_local(fst)
            fst_minimize_encoded(fst)
            fst_add_self_loops(
                fst, self.transition_model, disambig_syms_in, self.options.self_loop_scale
            )
        else:
            transcript_symbols = [self.to_int(x) for x in transcript.split()]
            fst = self.compiler.CompileGraphFromText(transcript_symbols)
        if fst.Start() == pywrapfst.NO_STATE_ID:
            logger.warning(f"Could not construct FST for '{transcript}")
            return None
        return fst
